// SPDX-License-Identifier: MIT
pragma solidity >=0.8.0 <0.9.0;

import "@openzeppelin/contracts/utils/cryptography/ECDSA.sol";
import "./Utils.sol";

library SignatureUtils {
    // Utility function to verify geth style signatures
    function verifySig(
        address _signer,
        bytes32 _messageDigest,
        Utils.Signature calldata _sig
    ) private pure returns (bool) {
        return _signer == ECDSA.recover(_messageDigest, _sig.v, _sig.r, _sig.s);
    }

    function checkValidatorSignatures(
        // The current validator set and their powers
        Utils.ValsetArgs calldata _currentValset,
        // The current validator's signatures
        Utils.Signature[] calldata _sigs,
        // This is what we are checking they have signed
        bytes32 _theHash,
        uint64 _powerThreshold
    ) internal pure {
        uint64 cumulativePower = 0;

        for (uint64 i = 0; i < _currentValset.validators.length; i++) {
            // If v is set to 0, this signifies that it was not possible to get a signature
            // from this validator and we skip evaluation
            // (In a valid signature, it is either 27 or 28)
            if (_sigs[i].v != 0) {
                // Check that the current validator has signed off on the hash
                if (!verifySig(_currentValset.validators[i], _theHash, _sigs[i])) {
                    revert("Utils.InvalidSignature()");
                }

                // Sum up cumulative power
                cumulativePower = cumulativePower + _currentValset.powers[i];

                // Break early to avoid wasting gas
                if (cumulativePower > _powerThreshold) {
                    break;
                }
            }
        }

        // Check that there was enough power
        if (cumulativePower <= _powerThreshold) {
            revert("Utils.InsufficientPower(cumulativePower, _powerThreshold)");
        }
        // Success
    }
}
